-- Puffin: A Tojaso FowlPlay Project
-- Widget class for custom nameplates with a bar plus texts, panels, icons, and auras.
-- Nameplates are not recyclable and assume that once allocated they are used forever.
-- Attributes are generally same as for each base class except contained in subtables.
-- Most attributes are considered static and only ones identified with * can be changed.

local _, Puffin = ...
local EF = Puffin.Error
local media = LibStub("LibSharedMedia-3.0")
local Align = Puffin.Align
local Pixel = Puffin.Pixel
local class = Puffin.Class
local Widget_Backdrop = Puffin.Classes.Widget_Backdrop

-- Widget to display nameplates, subclass of Widget_Backdrop

-- Attributes that apply to the overall nameplate:
-- width, height - dimensions of the widget's container that is parent frame for display elements
-- scale - factor by which to increase/decrease the size of the widget
-- strata - frame strata for the widget
-- level - base frame level for the widget
-- depth - relative frame level for the widget (higher shows on top of lower)
-- alpha - factor by which to increase/decrease opacity of the widget
-- hidden - boolean to toggle hiding the widget
-- condition - boolean that results in hiding the widget when exactly equal to false
-- bar - attribute table for the bar
-- texts - subtables for each text, indexed by number
-- icons - subtables for each icon, indexed by number
-- panels - subtables for each panel, indexed by number
-- auras - subtables for each aura, indexed by number

-- Bar attributes:
-- fgTexture - texture to use for the bar foreground (default: color swatch)
-- fgColor -- color to use for the bar foreground (default: blue) *
-- fgAlpha - opacity of the bar foreground (default: 1)
-- bgTexture - texture to use for the bar background (default: same as foreground)
-- bgColor - color to use for the bar background (default: same as foreground)
-- bgAlpha - opacity of the bar background (default: 0.6)
-- sparkAlpha - opacity of spark to show at end of bar (default: 0.5)
-- value - if defined then is the fraction of bar's foreground to fill and timer values are ignored (default: 1) *

-- Text attributes:
-- font - font to use for the text (default: ChatFontNormal)
-- fsize - font size for the text (default: 10)
-- delta - delta to apply to font size (default: 0)
-- shadow - enable a shadow behind the text (default: false)
-- outline - enable font "OUTLINE" flag (default: false)
-- thick - enable font "THICKOUTLINE" flag (default: false)
-- mono - enable font "MONOCHROME" flag (default: false)
-- wrap - enable wrap of text to multiple lines (default: false)
-- justifyH - horizontal text alignment (default: "CENTER")
-- justifyV - vertical text alignment (default: "MIDDLE")
-- text - string to display when textFormat is "text" (default: "") *
-- color - color to apply to text (default: none) *
-- width, height - dimensions of the text
-- layer - draw layer for the text
-- depth - relative depth within the layer for the text
-- point - point on text for setting position
-- relative - point on base frame for setting position
-- offsetX - horizontal displacement of text (default: 0)
-- offsetY - vertical displacement of text (default: 0)
-- alpha - factor by which to increase/decrease opacity of the text
-- condition - boolean that results in hiding the text when exactly equal to false *

-- Icon attributes (note these do not include backdrop):
-- texture - path to icons's texture (default: color swatch) *
-- texcoord - table containing icon's texcoord settings (default: { 0, 1, 0, 1 }) *
-- texcoordIgnore - don't set icon's texcoords if this is true (default: false) *
-- color - color to apply to icon (default: none) *
-- width, height - dimensions of the icon
-- layer - draw layer for the text
-- depth - relative depth within the layer for the text
-- point - point on icon for setting position
-- relative - point on base frame for setting position
-- offsetX - horizontal displacement of icon (default: 0)
-- offsetY - vertical displacement of icon (default: 0)
-- alpha - factor by which to increase/decrease opacity of the icon
-- condition - boolean that results in hiding the icon when exactly equal to false *

-- Panel attributes:
-- backdropTexture - if defined then show a background texture
-- backdropColor - color to shade the backdrop texture *
-- backdropTile - if defined then size of tile in pixels for repeating background texture
-- backdropInset - distance from edge of frame to backdrop texture (default: 2)
-- borderTexture - if defined then show a border around the widget
-- borderColor - color to apply to border (default: black) *
-- borderWidth - width of border (default: 8)
-- borderOffset - distance of border from widget's edge (default: 2)
-- width, height - dimensions of the panel
-- depth - relative frame level for the panel
-- point - point on panel for setting position
-- relative - point on base frame for setting position
-- offsetX - horizontal displacement of panel (default: 0)
-- offsetY - vertical displacement of panel (default: 0)
-- alpha - factor by which to increase/decrease opacity of the panel
-- condition - boolean that results in hiding the panel when exactly equal to false *

-- Aura attributes:
-- TBD - icon attributes
-- TBD - backdrop attributes
-- TBD - time text attributes
-- width, height - dimensions of the icon
-- depth - relative frame level for the icon
-- point - point on icon for setting position
-- relative - point on base frame for setting position
-- offsetX - horizontal displacement of icon (default: 0)
-- offsetY - vertical displacement of icon (default: 0)
-- alpha - factor by which to increase/decrease opacity of the icon
-- condition - boolean that results in hiding the icon when exactly equal to false *

local Widget_Nameplate = Widget_Backdrop:subclass("Widget_Nameplate")
Puffin.Classes.Widget_Nameplate = Widget_Nameplate

local nameplateNumber = 0

-- Create a new nameplate object
function Widget_Nameplate:initialize(template, layout)
	local nameplate = {}
	nameplateNumber = nameplateNumber + 1
	local frameName = "Puffin_Nameplate_Widget" .. tostring(nameplateNumber) -- generate unique name for nameplate's frame
	nameplate.container = CreateFrame("Frame", frameName .. "Container", UIParent)
	Widget_Backdrop.initialize(self, nameplate, template, layout) -- pass table of display frames used by this widget plus optional template and layout

	local t = self._display; assert(t, EF("Widget_Nameplate", "undefined widget"))
	local tc = t.container; assert(tc, EF("Widget_Nameplate", "undefined container"))
	
	if self.bar then -- create and configure the bar
		t.bgTexture = tc:CreateTexture(nil, "BACKGROUND")
		t.fgTexture = tc:CreateTexture(nil, "BORDER")	
		t.spark = tc:CreateTexture(nil, "OVERLAY")
		t.spark:SetTexture([[Interface\CastingBar\UI-CastingBar-Spark]])
		t.spark:SetBlendMode("ADD"); t.spark:SetTexCoord(0, 1, 0, 1)
		
		local tfg, tbg, tsp = t.fgTexture, t.bgTexture, t.spark
		local at = self.bar -- table of attributes
		local w, h = Pixel(at.width), Pixel(at.height)	
		local tex = at.bgTexture or at.fgTexture
		local c = at.bgColor or at.fgColor
		if tex then mt = media:Fetch("statusbar", tex, true); if mt then tex = mt end end -- check if background texture is in LibSharedMedia
		if tex then tbg:SetTexture(tex) else tbg:SetTexture(1, 1, 1) end
		if not c then c = defaultBlue end; tbg:SetVertexColor(c.r, c.g, c.b, 1)
		tbg:SetSize(w, h); tbg:SetTexCoord(0, 1, 0, 1); tbg:SetAlpha(at.bgAlpha or 0.6)
		tbg:ClearAllPoints(); tbg:SetPoint("TOP", tc, "TOP")

		if at.bgTexture ~= at.fgTexture then -- if same then already have texture for foreground
			tex = at.fgTexture -- if not same then look up foreground texture
			if tex then mt = media:Fetch("statusbar", tex, true); if mt then tex = mt end end -- check if foreground texture is in LibSharedMedia
		end
		if tex then tfg:SetTexture(tex) else tfg:SetTexture(1, 1, 1) end
		tfg:SetAlpha(at.fgAlpha or 1)
		tsp:SetSize(Pixel(16), h * 1.5); tsp:SetAlpha(at.sparkAlpha or 0)
	end
	
	if self.texts then -- create and configure texts based on attributes in subtable
		t.texts = Puffin.AllocateTable()
		for k, at in pairs(self.texts) do
			local w, h = Pixel(at.width), Pixel(at.height)	
			local fs = tc:CreateFontString(nil, "ARTWORK", ChatFontNormal)		
			t.texts[k] = fs
			fs:SetJustifyH(at.justifyH or "CENTER"); fs:SetJustifyV(at.justifyV or "MIDDLE"); fs:SetWordWrap(at.wrap)
			fs:SetShadowColor(0, 0, 0, at.shadow and 1 or 0); fs:SetShadowOffset(1, -1); fs:SetFontObject(ChatFontNormal)
			local font = at.font
			if font then local mf = media:Fetch("font", font, true); if mf then font = mf end end -- check if text font is in LibSharedMedia
			local fsize = (at.fsize or 10) + (at.delta or 0)
			if font then fs:SetFont(font, fsize, Puffin.TextFlags(at.outline, at.thick, at.mono)) end
			fs:SetSize(w, h)
			fs:ClearAllPoints(); fs:SetPoint(at.point, tc, at.relative, Pixel(at.offsetX or 0), Pixel(at.offsetY or 0))
			fs:SetAlpha(at.alpha or 1); fs:SetDrawLayer(at.layer or "ARTWORK", at.depth)
		end
	end

	if self.icons then -- create and configure icons based on attributes in subtable
		t.icons = Puffin.AllocateTable()
		for k, at in pairs(self.icons) do
			local w, h = Pixel(at.width), Pixel(at.height)	
			local tex = tc:CreateTexture(nil, "ARTWORK")		
			t.icons[k] = tex
			tex:SetSize(w, h)
			tex:ClearAllPoints(); tex:SetPoint(at.point, tc, at.relative, Pixel(at.offsetX or 0), Pixel(at.offsetY or 0))
			tex:SetAlpha(at.alpha or 1); tex:SetDrawLayer(at.layer or "ARTWORK", at.depth)
		end
	end
	
	if self.panels then -- create and configure panels based on attributes in subtable
		t.panels = Puffin.AllocateTable()
		for k, at in pairs(self.panels) do
			local w, h = Pixel(at.width), Pixel(at.height)
			local panel = CreateFrame("Frame", nil, tc)
			t.panels[k] = panel
			local b = { tile = false, tileSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0 }}
			assert(at.backdropTexture or at.borderTexture, EF("Widget_Nameplate", "panel needs backdrop or border")) -- check if showing a backdrop and/or border
			local tex = at.borderTexture
			if tex then mt = media:Fetch("border", tex, true); if mt or tex == "None" then tex = mt end end -- check if border texture is in LibSharedMedia
			b.edgeFile = tex
			tex = at.backdropTexture
			if tex then mt = media:Fetch("background", tex, true); if mt or tex == "None" then tex = mt end end -- check if background texture is in LibSharedMedia
			b.bgFile = tex
			local x = Pixel(at.backdropInset or 2)
			local ins = b.insets; ins.left = x; ins.right = x; ins.top = x; ins.bottom = x
			b.edgeSize = Pixel(at.borderWidth or 8); if (b.edgeSize < Pixel(1)) then b.edgeSize = Pixel(1) end
			if at.backdropTile and at.backdropTile > 0 then b.tile = true; b.tileSize = at.backdropTile else b.tile = false; b.tileSize = 0 end
			local offset = Pixel(2 * (at.borderOffset or 2))
			panel:SetSize(w + offset, h + offset)
			panel:SetBackdrop(b)
			panel:ClearAllPoints(); panel:SetPoint(at.point, tc, at.relative, Pixel(at.offsetX or 0), Pixel(at.offsetY or 0))
			panel:SetAlpha(at.alpha or 1); panel:SetFrameLevel(tc:GetFrameLevel() + at.depth)
		end
	end
	
	if self.auras then -- create and configure auras based on attributes in subtable
		t.auras = Puffin.AllocateTable()
		local at = self.auras -- attributes are shared across all auras
		local w, h = Pixel(at.width), Pixel(at.height)
		local offset = Pixel(2 * (at.borderOffset or 2))
		local tex = at.borderTexture
		assert(tex, EF("Widget_Nameplate", "aura needs border")) -- check for required border texture
		if tex then mt = media:Fetch("border", tex, true); if mt or tex == "None" then tex = mt end end -- check if border texture is in LibSharedMedia
		local b = { edgeFile = tex, tile = false, tileSize = 0, insets = { left = 0, right = 0, top = 0, bottom = 0 }}
		b.edgeSize = Pixel(at.borderWidth or 8); if (b.edgeSize < Pixel(1)) then b.edgeSize = Pixel(1) end
		local c = at.borderColor or defaultBlack
		local font = at.font
		if font then local mf = media:Fetch("font", font, true); if mf then font = mf end end -- check if text font is in LibSharedMedia
		local fsize = (at.fsize or 10) + (at.delta or 0)
		local fflags = Puffin.TextFlags(at.outline, at.thick, at.mono)
		local tpos = at.timePosition
		assert(tpos, EF("Widget_Nameplate", "aura needs border")) -- check for required border texture
		for i = 1, at.maxAuras do -- create and configure each aura using shared attributes
			local aura = CreateFrame("Frame", nil, tc) -- first, create the base frame for the aura
			t.auras[i] = aura
			aura:SetSize(w + offset, h + offset)
			aura:SetBackdrop(b) -- and add the backdrop to the base frame
			local pt = at.frames[i] -- position info is in numbered subtables
			aura:ClearAllPoints(); aura:SetPoint(pt.point, tc, pt.relative, Pixel(pt.offsetX or 0), Pixel(pt.offsetY or 0))
			aura:SetAlpha(at.alpha or 1); aura:SetFrameLevel(tc:GetFrameLevel() + at.depth)
			aura:SetBackdropBorderColor(c.r, c.g, c.b, c.a or 1)
			
			local tex = tc:CreateTexture(nil, "ARTWORK") -- second, create the icon texture for the aura
			aura._icon = tex -- and save a reference to it in the frame itself
			tex:SetSize(w, h)
			tex:SetTexCoord(0.07, 0.93, 0.07, 0.93) -- trim removes 7% of edges
			tex:ClearAllPoints(); tex:SetPoint("CENTER", aura, "CENTER", 0, 0)
			
			local fs = tc:CreateFontString(nil, "ARTWORK", ChatFontNormal) -- third, create the font string for the aura timer	
			aura._time = fs -- and save a reference to it in the frame itself
			fs:SetJustifyH(at.justifyH or "CENTER"); fs:SetJustifyV(at.justifyV or "MIDDLE"); fs:SetWordWrap(at.wrap)
			fs:SetShadowColor(0, 0, 0, at.shadow and 1 or 0); fs:SetShadowOffset(1, -1); fs:SetFontObject(ChatFontNormal)
			if font then fs:SetFont(font, fsize, fflags) end
			fs:SetSize(w, fsize + 2) -- scale height with font size
			fs:ClearAllPoints(); fs:SetPoint(tpos.point, aura, tpos.relative, Pixel(tpos.offsetX or 0), Pixel(tpos.offsetY or 0))
		end
	end
end

-- Update the widget's display objects based on current data attribute values (or defaults)
function Widget_Nameplate:Update()
	Widget_Backdrop.Update(self)
	local t = self._display; assert(t, EF("Widget_Rectangle", "undefined rect"))
	local tc = t.container; assert(tc, EF("Widget_Rectangle", "undefined container"))
	local showing = tc:IsShown()

	if self.bar then -- update bar
		local tfg, tbg, tsp = t.fgTexture, t.bgTexture, t.spark
		tc:Hide(); tfg:Hide(); tbg:Hide(); tsp:Hide()
		if showing then
			local at = self.bar -- attributes for the bar
			local w, h = Pixel(at.width), Pixel(at.height)	
			local fill = at.value or 1 -- value bars specify a fraction to show
			local fw = w * fill -- doesn't matter if absolutely exact
			if fw >= 0.25 then -- but must be aligned prior to this test!
				tfg:SetSize(fw, h); tfg:SetTexCoord(0, 0, 0, 1, fill, 0, fill, 1)
				tfg:ClearAllPoints(); tfg:SetPoint("TOPLEFT", tc, "TOPLEFT"); tfg:Show()
				tsp:ClearAllPoints(); tsp:SetPoint("CENTER", tfg, "RIGHT", 0, Pixel(-1))
				if at.sparkAlpha and at.sparkAlpha > 0 and fill < 1 then tsp:Show() else tsp:Hide() end
			else tfg:Hide(); tsp:Hide() end
			if at.fgColor then local c = at.fgColor; tfg:SetVertexColor(c.r, c.g, c.b, 1) end
			if at.bgAlpha ~= 0 then tbg:SetSize(w, h); tbg:Show() end
			if t.backdrop then
				local offset = Pixel(2 * (self.borderOffset or 2))
				t.backdrop:SetSize(w + offset, h + offset)
			end
			tc:SetSize(w, h); tc:Show()
		end
	end
	
	if self.texts then
		for k, at in pairs(self.texts) do -- update texts
			local fs = t.texts[k]
			fs:Hide()
			if showing and at.condition ~= false then
				local c = at.color; if c then fs:SetTextColor(c.r, c.g, c.b, 1) else fs:SetTextColor(0, 0, 0, 1) end
				fs:SetText(at.text)
				fs:Show()
			end
		end
	end
	
	if self.icons then
		for k, at in pairs(self.icons) do -- update icons
			local tex = t.icons[k]
			tex:Hide()
			if showing and at.condition ~= false then
				if at.texture then tex:SetTexture(at.texture) else tex:SetTexture(1, 1, 1) end
				if at.raidIndex then -- special case for raid target icon!
					SetRaidTargetIconTexture(tex, at.raidIndex)
				else
					if not at.texcoordIgnore then
						if at.texcoord then tex:SetTexCoord(unpack(at.texcoord)) else tex:SetTexCoord(0, 1, 0, 1) end
					end
					local c = at.color; if c then tex:SetVertexColor(c.r, c.g, c.b) else tex:SetVertexColor(1, 1, 1) end		
				end
				tex:Show()
			end
		end
	end

	if self.panels then
		for k, at in pairs(self.panels) do -- update panels
			local panel = t.panels[k]
			panel:Hide()
			if showing and at.condition ~= false then
				local c = at.backdropColor or defaultWhite; panel:SetBackdropColor(c.r, c.g, c.b, c.a or 1)			
				c = at.borderColor or defaultBlack; panel:SetBackdropBorderColor(c.r, c.g, c.b, c.a or 1)
				panel:ClearAllPoints(); panel:SetPoint(at.point, tc, at.relative, Pixel(at.offsetX or 0), Pixel(at.offsetY or 0))
				panel:Show()
			end
		end
	end
	
	if self.auras then
		local at = self.auras -- attributes are shared across all auras
		for i = 1, at.maxAuras do
			local aura = t.auras[i]
			aura:Hide(); aura._icon:Hide(); aura._time:Hide()
			if showing and i <= at.activeAuras then
				local pt = at.frames[i]
				if pt.texture then
					aura._icon:SetTexture(pt.texture) -- set the icon for the aura
					aura._icon:Show()
				end
				if pt.timeLeft then
					local ts = Puffin.FormatTime(pt.timeLeft, at.timeFormat, false, false)
					aura._time:SetText(ts) -- set the time remaining as an integer
					aura._time:Show()
				end
				aura:Show()
			end
		end
	end
end
